/**
 * Nextion-Calculator.ino - A calculator similar to the Windows 10 standard calculator.
 * 
 * Copyright (C) 2020 Mario Gianota (gianotamario@gmail.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *---------------------------------------------------------------------------------
 *
 * A useful calculator for the Arduino Uno. The calculator is similar in style to the
 * standard calculator that ships with Windows 10. Note: It does not include the
 * scientific and programmer functions that the Windows 10 calculator does, but these
 * functions may be implemented at a later date.
 * 
 * The calculator provides a set of 10 functions:
 * 
 * - Add, Subtract, Multiply, Divide
 * - Percentage calculation
 * - 1/x calculation
 * - Square root
 * - Square
 * - [C]ancel - clears the calculator's memory
 * - [CE] Clear Entry - Deletes the last entry made into the calculator
 * 
 * All calculations are carried out with double precision. Note that because of the Arduino
 * being as small as it is, the output of decimals is limited to *two* places.
 * 
 * This version of the calculator uses a Nextion NX4832T035 3.5" HMI TFT LCD Display
 * which should be connected to the TX/RX pins of the Arduino. Since all communication
 * to and from the LCD is done via UART you can upload this sketch to the Arduino and
 * use the Arduino IDE's serial monitor to interact with it.
 *
 * See the README.md file for hardware setup.
 * 
 * HACKING:
 * The calculator is built in a classic style with a scanner that gathers sequences of
 * characters into tokens and a top-down, recursive descent parser that processes the tokens.
 * Expression evaluation is done with a stack.
 */

#define MAX_TOKENS 64     // Maximum size of an expression in tokens
#define MAX_STACK 100     // Maximum size of evaluation stack

enum TokenType {NUMBER, ADD, MULTIPLY, SUBTRACT, DIVIDE, EQUALS, CE, C, ONE_OVER, SQRT,
                PERCENT, SQUARE, EOL};

char* typeNames[] = {"Number", "+", "*", "-", "/", "=", "CE", "C", "1/x", "Sqrt", "%", "Sqr", "Eol"};

class Token {
  public:
    TokenType type;
    double value;
    Token() {}
    Token(TokenType t, double v) {
      type = t;
      value = v;  
    }
};

Token* tokens[MAX_TOKENS];// Scanned Token objects
int tokenIndex;           // Current token index
Token* lookToken;         // Lookahead token for parser

char look;                // Lookahead character for scanner
int charIndex;            // Current scanning position

String expr = "";         // The expression entered by the user to scan, parse & evaluate
String buf = "";          // Temporary buffer for parsing numbers

double output;            // Output result

bool error;               // error flag - set if the entered expression is malformed

// Evaluation stack
class Stack {
  int top;
  public:
  double myStack[MAX_STACK]; //stack array
  
  Stack() { top = -1; }
  bool push(double x);
  double pop();
  bool isEmpty();
  void reset();
};

//pushes element on to the stack
bool Stack::push(double item) {
  if (top >= (MAX_STACK-1)) {
    //Stack Overflow
    //Serial.println("push() Stack overflow!");
    return false;
 } else {
   myStack[++top] = item;
   return true;
 }
}
  
//removes or pops elements out of the stack
double Stack::pop() {
   if (top < 0) {
      //Stack Underflow
      //Serial.println("pop() Stack underflow!");
      return 0;
   } else {
      double item = myStack[top--];
      return item;
   }
}
  
//check if stack is empty
bool Stack::isEmpty() {
   return (top < 0);
}  

// clear the stack
void Stack::reset() {
  top = -1;
}

Stack stack;      // Evaluation stack
bool entryStarted = true;   // Flag set when an entry is first begun

void setup() {
  Serial.begin(9600);
  while(!Serial)
    ;  

}

void loop() {
  char c = 0;
  entryStarted = expr.length() == 0;

  // Read a character from the serial port.
  if( Serial.available() > 0 ) {
    c = (char)Serial.read();
    if( c == 'E' ) {
      // Clear last entry command
      if( expr.length() >= 1 ) {
        expr.remove(expr.length()-1,2);
        Serial.print("t0.txt=\"");
        Serial.print(expr);
        Serial.print("\"");
        Serial.write(0xFF);
        Serial.write(0xFF);
        Serial.write(0xFF);
      }
    } else if( c == 'C' ) {
      // clear all input command
      expr.remove(0, expr.length());
      output = 0;
      String s2(output);
      s2 = s2.substring(0, s2.length() - 3);
      Serial.print("t0.txt=\"");
      Serial.print(s2);
      Serial.print("\"");
      Serial.write(0xFF);
      Serial.write(0xFF);
      Serial.write(0xFF);
    } else {

      // If the entry has just begun and one of the basic operators is entered then
      // insert the value of the output into the expression so that the user can
      // do a running total
      if( entryStarted && (c == '+' || c == '-' || c == '/' || c == '*' || c == 'S' || c == 'Q' || c == 'O' || c == '%')) {
        String s3(output);
        if( s3.endsWith(".00") )
          s3 = s3.substring(0, s3.length() - 3);
        expr.concat(s3);    
      }

      // Check length. We can only display 11 chars in the UI
      //if( expr.length() <= 11 )
        expr.concat(c); // Add the char to the expression string & output expression
      entryStarted = false;
      
      if( c != '=' || c !='O' || c != 'S' || c != 'Q' ) {
        Serial.print("t0.txt=\"");
        Serial.print(expr);
        Serial.print("\"");
        Serial.write(0xFF);
        Serial.write(0xFF);
        Serial.write(0xFF);
      }
    }
  }

  // If the character is any one of those that will trigger a scan & calculation of the expression, then do it
  if( c == '=' || c == 'O' || c == 'S' || c == 'Q' ) { // equals, one over x, square root & square trigger a scan and calculation
    scan();
    // output tokens
    for(int i=0; i<tokenIndex; i++) {
      Token* t = tokens[i];
      //Serial.print( typeNames[t->type] );
      //Serial.print(" ");
      //Serial.print( t->value );
      //Serial.print(",");
    }

    calculate();

    // reset expr
    expr.remove(0, expr.length());
    
    // If we don't have an error, then send the output
    if( ! error ) {
      Serial.print("t0.txt=\"");
      String s(output);
      if( s.endsWith(".00") )
        s = s.substring(0, s.length() - 3);
        
      // Check the length doesn't overflow the textfield UI
      if( s.length() > 11 ) {
        s = s.substring(0, 11);
      }
      
      Serial.print(s);
      Serial.print("\"");
      Serial.write(0xFF);
      Serial.write(0xFF);
      Serial.write(0xFF);
      
    } else {
      Serial.print("t0.txt=\"Error\"");
      Serial.write(0xFF);
      Serial.write(0xFF);
      Serial.write(0xFF);
      
    }
    error = false;
  }
  delay(100);
}


// Breaks up the expression into an array of Token objects
void scan() {
  
  // Free memory 
  for(int i=0; i<tokenIndex; i++) {
    delete tokens[i];
  }

  // reset token and expr indices
  tokenIndex = 0;
  charIndex = 0;

  // prime the pump with the first character
  nextChar();

  // scan the expression string and create Tokens
  while( look != '\0' ) {
    Token* t;

    switch(look) {
      case '.':
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        t = new Token(NUMBER, getNumber());
        tokens[tokenIndex++] = t;
      break;
      case '+':
        t = new Token(ADD, 0);
        tokens[tokenIndex++] = t;
        nextChar();
      break;
      case '-':
        t = new Token(SUBTRACT, 0);
        tokens[tokenIndex++] = t;
        nextChar();
      break;
      case '*':
        t = new Token(MULTIPLY, 0);
        tokens[tokenIndex++] = t;
        nextChar();
      break;
      case '/':
        t = new Token(DIVIDE, 0);
        tokens[tokenIndex++] = t;
        nextChar();
      break;
      case '=':
        t = new Token(EQUALS, 0);
        tokens[tokenIndex++] = t;
        nextChar();
        break;
      case 'O': // ONE_OVER
        t = new Token(ONE_OVER, 0);
        tokens[tokenIndex++] = t;
        nextChar();
        break;
      case 'S': // SQRT
        t = new Token(SQRT, 0);
        tokens[tokenIndex++] = t;
        nextChar();
        break;
      case '%': // PERCENT
        t = new Token(PERCENT, 0);
        tokens[tokenIndex++] = t;
        
        nextChar();
        break;
      case 'Q': // SQUARE
        t = new Token(SQUARE, 0);
        tokens[tokenIndex++] = t;
        nextChar();
        break;
      default:
        // We should never get here, unless someone is driving the sketch from the serial monitor
        // and enters an unrecognized character
        nextChar();
    }
  }
  // Create EOL token to mark the end of the token list. Note: Not really neccessary.
  Token* t2 = new Token(EOL, 0);
  tokens[tokenIndex++] = t2;
}

// Read next character from expression
void nextChar() {
  if( charIndex >= expr.length() ) {
    look = '\0';  
    return;
  }
  look = expr.charAt(charIndex++);
}

// parse a number
double getNumber() {
  // clear the buffer
  buf.remove(0, buf.length());

  // hack to deal with the case where someone enters a decimal
  // without preceding it with 0.
  if( look == '.' ) {
    buf.concat("0.");
    nextChar();
  }
  
  while( isPartNumber(look) ) {
   buf.concat(look);
   nextChar();
  }
  return buf.toDouble(); // Gotta love Arduino for providing a double parser, limited as it is.
}

// return true if c is a digit, or a decimal point
bool isPartNumber(char c) {
  if( c == '.' || c == '0' || c == '1' || c == '2' || c == '3'
      || c == '4' || c == '5' || c == '6' || c == '7' || c == '8'
      || c == '9' )
      return true;
  return  false;
}

// Calculate the result of an expression
void calculate() {
  tokenIndex = 0;
  lookToken = NULL;
  nextToken();
  stack.reset();

  // parse and calculate
  expression();
  
  if( !error ) {
    output = stack.pop(); // result is left as the top item on the stack

    // Do we have a function?
    if( lookToken->type == SQUARE )
      output = output * output;
    else if( lookToken->type == ONE_OVER ) {
      if( output != 0.0 ) // check for division by zero
        output = 1/output;
      else
        error = 1;
    } else if( lookToken->type == SQRT ) 
      output = sqrt(output);
  }
}

void nextToken() {
  
  if( lookToken->type == EOL ) {
    //Serial.println("EOL");
    return;
  }
    
  lookToken = tokens[tokenIndex++];
}

// parse and calculate the result of an expression
// Expression := EOL | [UnaryOp] <term> [<AddOp> <number>]*
void expression() {
  
  if( lookToken->type == EOL ) 
    return;
  
  if( isAddOp(lookToken->type ) ){
    stack.push(0);
  } else
    term();
  
  while( isAddOp(lookToken->type ) ) {
    if( lookToken->type == ADD ) {
      add();  
    } else if( lookToken->type == SUBTRACT ) {
      subtract();
    }
  }
}

// Parse and calculate a term
// <term> := <factor> [ <multiply> <factor> ]*
void term() {
  factor();

  while( isMulOp(lookToken->type) ) {
    if( lookToken->type == MULTIPLY )
      multiply();
    else if( lookToken->type == DIVIDE )
      divide();
    else if( lookToken->type == PERCENT )
      percent();  
  }
  
}

void factor() {
  // check for unary -, +
  bool negate = false;
  if( isAddOp(lookToken->type) ) {
    if( lookToken->type == ADD ) {
      nextToken();
    } else {
      negate = true;
      nextToken();
    }
  }
  if( negate ) {
    lookToken->value *= -1;
  }
  if( lookToken->type != NUMBER ) {
    error = true;
  }
  stack.push(lookToken->value);
  nextToken();
}

void add() {
  nextToken();
  term();
  double right = stack.pop();
  double left = stack.pop();
  stack.push(left + right);
}

void subtract() {
  nextToken();
  term();
  double right = stack.pop();
  double left = stack.pop();
  stack.push(left - right);
}

void multiply() {
  nextToken();
  term();
  double right = stack.pop();
  double left = stack.pop();
  stack.push(left * right);
}

void divide() {
  nextToken();  
  term();
  double right = stack.pop();
  double left = stack.pop();
  stack.push(left / right);
}

void percent() {
  nextToken();
  double pd = stack.pop() / 100.0;
  term();
  stack.push(pd * stack.pop());
}


bool isAddOp(int tokenType) {
  return tokenType == ADD || tokenType == SUBTRACT;
}

bool isMulOp(int tokenType) {
  return tokenType == MULTIPLY || tokenType == DIVIDE || tokenType == PERCENT;
}
